/*
 * OLSR ad-hoc routing table management protocol
 * Copyright (C) 2003 Andreas Tnnesen (andreto@ifi.uio.no)
 *
 * This file is part of olsrd-unik.
 *
 * olsrd-unik is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * olsrd-unik is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with olsrd-unik; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "defs.h"
#include "interface.h"
#include <net/if.h>
#include <net/if_arp.h>
#include <asm/types.h>
#include <arpa/inet.h>
#include <netdb.h>


/**
 *Set up interfaces to use.
 *
 *@return the number of interfaces configured
 */
int
ifinit(struct if_name *ifnames)
{
  struct interface ifs, *ifp;
  struct if_name *tmp_if;
  int s;
  struct ifreq *ifr;
  int nbinterf;

  nbinterf = 0;

  /*
   *socket for icotl call
   */
  if ((s = socket(ipversion, SOCK_DGRAM, 0)) < 0) 
    {
      syslog(LOG_ERR, "socket: %m");
      close(s);
      return -1;
    }

  ifr = olsr_malloc(sizeof(struct ifreq), "Interface 1");

  olsr_printf(1, "\n ---- Interface configuration ---- \n\n");
  for(tmp_if = ifnames; tmp_if != NULL; tmp_if = tmp_if->next)
    {

      memset(ifr, 0, sizeof(struct ifreq));
      strncpy(ifr->ifr_name, tmp_if->name, IFNAMSIZ);
      olsr_printf(1, "%s:\n", tmp_if->name);

      /* Get flags (and check if interface exists) */
      if (ioctl(s, SIOCGIFFLAGS, ifr) < 0) 
	{
	  olsr_printf(1, "\tNo such interface!\n");
	  continue;
	}

      ifs.int_flags = ifr->ifr_flags | IFF_INTERFACE;

      olsr_printf(3, "\tFlags: %08x\n", ifs.int_flags);
      


      if ((ifs.int_flags & IFF_UP) == 0)
	{
	  olsr_printf(1, "\tInterface not up - skipping it...\n");
	  continue;
	}

      /* Check broadcast */
      if ((ipversion == AF_INET) && (!(ifs.int_flags & IFF_BROADCAST))) 
	{
	  olsr_printf(1, "\tNo broadcast - skipping\n");
	  continue;
	}

      if (ifs.int_flags & IFF_POINTOPOINT)
	{
	  olsr_printf(1, "\tThis is a point-to-point interface - skipping it...\n");
	  continue;
	}


      if (ifs.int_flags & IFF_LOOPBACK)
	{
	  olsr_printf(1, "\tThis is a loopback interface - skipping it...\n");
	  continue;
	}

      /* trying to detect if interface is wireless. */
      if(check_wireless_interface(s, ifr))
	{
	  olsr_printf(1, "\tWireless interface detected\n");
	  ifs.is_wireless = 1;
	}
      else
	{
	  olsr_printf(1, "\tNot a wireless interface\n");
	  ifs.is_wireless = 0;
	}


      /* Get interface index */

      ifs.if_index = if_nametoindex(ifr->ifr_name);

      /* IP version 6 */
      if(ipversion == AF_INET6)
	{
	  /* Get interface address */

	  if(get_ipv6_address(ifr->ifr_name, &ifs.int6_addr, ipv6_addrtype) <= 0)
	    {
	      if(ipv6_addrtype == IPV6_ADDR_SITELOCAL)
		olsr_printf(1, "\tCould not find site-local IPv6 address for %s\n", ifr->ifr_name);
	      else
		olsr_printf(1, "\tCould not find global IPv6 address for %s\n", ifr->ifr_name);
		
	      continue;
	    }

	  olsr_printf(1, "\tAddress: %s\n", ip6_to_string(&ifs.int6_addr.sin6_addr));
	  

	  /* Set default multicast address */
	  if(inet_pton(AF_INET6, ipv6_mult, &ifs.int6_multaddr.sin6_addr) < 0)
	    {
	      perror("Convert multicastaddr");
	      continue;
	    }
	  
	  /* Set address family */
	  ifs.int6_multaddr.sin6_family = AF_INET6;
	  /* Set port */
	  ifs.int6_multaddr.sin6_port = olsr_udp_port;

	  olsr_printf(1, "\tMulticast: %s\n", ip6_to_string(&ifs.int6_multaddr.sin6_addr));

	}
      /* IP version 4 */
      else
	{
	  /* Get interface address (IPv4)*/
	  if (ioctl(s, SIOCGIFADDR, ifr) < 0) 
	    {
	      olsr_printf(1, "\tCould not get address of interface - skipping it\n");
	      continue;
	    }
	  
	  ifs.int_addr = ifr->ifr_addr;
	  
	  olsr_printf(1, "\tAddress:%s\n", sockaddr_to_string(&ifs.int_addr));

	  /* Find netmask */
	  
	  if (ioctl(s, SIOCGIFNETMASK, ifr) < 0) 
	    {
	      syslog(LOG_ERR, "%s: ioctl (get broadaddr)", ifr->ifr_name);
	      continue;
	    }
	  
	  ifs.int_netmask = ifr->ifr_netmask;
	  
	  olsr_printf(3, "\tNetmask:%s\n", sockaddr_to_string(&ifs.int_netmask));

	  /* Find broadcast address */
	  
	  if (ioctl(s, SIOCGIFBRDADDR, ifr) < 0) 
	    {
	      syslog(LOG_ERR, "%s: ioctl (get broadaddr)", ifr->ifr_name);
	      continue;
	    }
	  
	  ifs.int_broadaddr = ifr->ifr_broadaddr;
	  
	  olsr_printf(1, "\tBroadcast address:%s\n", sockaddr_to_string(&ifs.int_broadaddr));

	  /* Deactivate IP spoof filter */
	  deactivate_spoof(ifr->ifr_name, nbinterf, ipversion);

	  /* Disable ICMP redirects */
	  disable_redirects(ifr->ifr_name, nbinterf, ipversion);

	}
      
      
      /* Set interface metric */
      ifs.int_metric = ifs.is_wireless;
	
      /* setting the interfaces number*/
      ifs.if_nr = nbinterf++; 
      olsr_printf(1, "\tInterface %s set up for use with index %d\n\n", tmp_if->name, ifs.if_nr);
      
      
      
      ifp = olsr_malloc(sizeof (struct interface), "Interface 2");
      
      memcpy(ifp, &ifs, sizeof(struct interface));
      
      ifp->int_name = olsr_malloc(strlen(ifr->ifr_name) + 1, "Interface 3");
      
      strcpy(ifp->int_name, ifr->ifr_name);
      /* Segfaults if using strncpy(IFNAMSIZ) why oh why?? */
      ifp->int_next = ifnet;
      ifnet = ifp;
	
    } /* END INTERFACE loop */
  
  free(ifr);
  close(s);

  return nbinterf;
}




/**
 *Find the local interface with a given address.
 *
 *@param addr the address to check.
 *
 *@return the interface struct representing the interface
 *that matched the address.
 */

struct interface *
if_ifwithaddr(union olsr_ip_addr *addr)
{
  struct interface *ifp;

  for (ifp = ifnet; ifp; ifp = ifp->int_next)
    {
      if(ipversion == AF_INET)
	{
	  /* IPv4 */
	  //printf("Checking: %s == ", inet_ntoa(((struct sockaddr_in *)&ifp->int_addr)->sin_addr));
	  //printf("%s\n", olsr_ip_to_string(addr));

	  if (COMP_IP(&((struct sockaddr_in *)&ifp->int_addr)->sin_addr, addr))
	      return ifp;
	}
      else
	{
	  /* IPv6 */
	  //printf("Checking %s ", olsr_ip_to_string((union olsr_ip_addr *)&ifp->int6_addr.sin6_addr));
	  //printf("== %s\n", olsr_ip_to_string((union olsr_ip_addr *)&((struct sockaddr_in6 *)addr)->sin6_addr));
	  if (COMP_IP(&ifp->int6_addr.sin6_addr, addr))
	    return ifp;
	}
    }
  return NULL;
}



/**
 *Find the interface with a given number.
 *
 *@param nr the number of the interface to find.
 *
 *@return return the interface struct representing the interface
 *that matched the number.
 */
struct interface *
if_ifwithsock(int fd)
{
  struct interface *ifp;
  ifp = ifnet;

  while (ifp) 
    {
      if (s[ifp->if_nr] == fd)
	return ifp;
      ifp = ifp->int_next;
    }
  
  return (ifp);
}


/*
 *From net-tools lib/interface.c
 *
 */

int
get_ipv6_address(char *ifname, struct sockaddr_in6 *saddr6, int scope_in)
{
  char addr6[40], devname[IFNAMSIZ];
  char addr6p[8][5];
  int plen, scope, dad_status, if_idx;
  FILE *f;
  struct sockaddr_in6 tmp_sockaddr6;

  if ((f = fopen(_PATH_PROCNET_IFINET6, "r")) != NULL) 
    {
      while (fscanf(f, "%4s%4s%4s%4s%4s%4s%4s%4s %02x %02x %02x %02x %20s\n",
		    addr6p[0], addr6p[1], addr6p[2], addr6p[3],
		    addr6p[4], addr6p[5], addr6p[6], addr6p[7],
		    &if_idx, &plen, &scope, &dad_status, devname) != EOF) 
	{
	  if (!strcmp(devname, ifname)) 
	    {
	      sprintf(addr6, "%s:%s:%s:%s:%s:%s:%s:%s",
		      addr6p[0], addr6p[1], addr6p[2], addr6p[3],
		      addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
	      if(debug_level > 1)
		{
		  olsr_printf(1, "\tinet6 addr: %s\n", addr6);
		  olsr_printf(3, "\tScope:");
		}
	      if(scope == scope_in)
		{
		  olsr_printf(3, "IPv6 addr:\n");
		  inet_pton(AF_INET6,addr6,&tmp_sockaddr6);
		  memcpy(&saddr6->sin6_addr, &tmp_sockaddr6, sizeof(struct in6_addr));	  
		  return 1;
		}
	    }
	}
      fclose(f);
    }
  
  return 0;
}




int
set_flag(char *ifname, short flag, int s)
{
  struct ifreq ifr;

  strncpy(ifr.ifr_name, ifname, IFNAMSIZ);

  /* Get flags */
  if (ioctl(s, SIOCGIFFLAGS, &ifr) < 0) 
    {
      fprintf(stderr,"ioctl (get interface flags)");
      return -1;
    }

  strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
  
  //printf("Setting flags for if \"%s\"\n", ifr.ifr_name);

  if(!(ifr.ifr_flags & IFF_UP & IFF_RUNNING))
    {
      /* Add UP */
      ifr.ifr_flags |= (IFF_UP | IFF_RUNNING);
      /* Set flags + UP */
      if(ioctl(s, SIOCSIFFLAGS, &ifr) < 0)
	{
	  fprintf(stderr, "ERROR(%s): %s\n", ifr.ifr_name, strerror(errno));
	  return -1;
	}
    }
  return 1;

}




/**
 *Create a new interf_name struct using a given
 *name and insert it into the interface list.
 *
 *@param name the name of the interface.
 *
 *@return nada
 */
void
queue_if(char *name)
{

  struct if_name *interf_n;

  //printf("Adding interface %s\n", name);

  /* check if the inerfaces already exists */
  interf_n = if_names;
  while(interf_n != NULL)
    {
      if(memcmp(interf_n->name, name, strlen(name)) == 0)
	{
	  fprintf(stderr, "Duplicate interfaces defined... not adding %s\n", name);
	  return;
	}
      interf_n = interf_n->next;
    }

  interf_n = olsr_malloc(sizeof(struct if_name), "queue interface");

  /* strlen () does not return length including terminating /0 */
  interf_n->name = olsr_malloc(strlen(name) + 1, "queue interface name");

  strcpy(interf_n->name, name);
  interf_n->next = if_names;
  if_names = interf_n;

}
